/*
 * scene.c - scene management
 * Copyright (C) 2008-2010  Alexandre Martins < alemartf [.a.t.] gmail [.d.o.t.] com >
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "scene.h"
#include "util.h"
#include "level.h"
#include "gameover.h"
#include "pause.h"
#include "quest.h"
#include "menu.h"
#include "intro.h"
#include "endofdemo.h"
#include "questover.h"
#include "confirmbox.h"
#include "langselect.h"
#include "credits.h"
#include "options.h"
#include "stageselect.h"


/* constants */
#define STORYBOARD_CAPACITY         256     /* up to 256 scenes in the storyboard */
#define SCENESTACK_CAPACITY         16      /* up to 16 scenes running simultanely */


/* private stuff */
static scene_t *storyboard[STORYBOARD_CAPACITY];
static scene_t *scene_create(void (*init_func)(), void (*update_func)(), void (*render_func)(), void (*release_func)());
static void scene_destroy(scene_t *scn);
static scene_t* scenestack[SCENESTACK_CAPACITY];
static int scenestack_size;



/* Storyboard */

/*
 * storyboard_init()
 * Initializes the storyboard
 */
void storyboard_init()
{
    int i;

    /* initializing... */
    for(i=0; i<STORYBOARD_CAPACITY; i++)
        storyboard[i] = NULL;

    /* registering the scenes */
    storyboard[SCENE_LEVEL] = scene_create(level_init, level_update, level_render, level_release);
    storyboard[SCENE_PAUSE] = scene_create(pause_init, pause_update, pause_render, pause_release);
    storyboard[SCENE_GAMEOVER] = scene_create(gameover_init, gameover_update, gameover_render, gameover_release);
    storyboard[SCENE_QUEST] = scene_create(quest_init, quest_update, quest_render, quest_release);
    storyboard[SCENE_MENU] = scene_create(menu_init, menu_update, menu_render, menu_release);
    storyboard[SCENE_INTRO] = scene_create(intro_init, intro_update, intro_render, intro_release);
    storyboard[SCENE_ENDOFDEMO] = scene_create(endofdemo_init, endofdemo_update, endofdemo_render, endofdemo_release);
    storyboard[SCENE_QUESTOVER] = scene_create(questover_init, questover_update, questover_render, questover_release);
    storyboard[SCENE_CONFIRMBOX] = scene_create(confirmbox_init, confirmbox_update, confirmbox_render, confirmbox_release);
    storyboard[SCENE_LANGSELECT] = scene_create(langselect_init, langselect_update, langselect_render, langselect_release);
    storyboard[SCENE_CREDITS] = scene_create(credits_init, credits_update, credits_render, credits_release);
    storyboard[SCENE_OPTIONS] = scene_create(options_init, options_update, options_render, options_release);
    storyboard[SCENE_STAGESELECT] = scene_create(stageselect_init, stageselect_update, stageselect_render, stageselect_release);
}




/*
 * storyboard_release()
 * Releases the storyboard
 */
void storyboard_release()
{
    int i;

    for(i=0; i<STORYBOARD_CAPACITY; i++) {
        if(storyboard[i])
            scene_destroy(storyboard[i]);
    }
}




/*
 * storyboard_get_scene()
 * Gets a scene from the storyboard.
 * Provide a SCENE_* value
 */
scene_t* storyboard_get_scene(int scene_id)
{
    scene_id = clip(scene_id, 0, STORYBOARD_CAPACITY-1);
    return storyboard[scene_id];
}




/* Scene stack */

/*
 * scenestack_init()
 * Initializes the scene stack
 */
void scenestack_init()
{
    int i;

    scenestack_size = 0;
    for(i=0; i<SCENESTACK_CAPACITY; i++)
        scenestack[i] = NULL;
}



/*
 * scenestack_release()
 * Releases the scene stack
 */
void scenestack_release()
{
    while(!scenestack_empty())
        scenestack_pop();
}



/*
 * scenestack_push()
 * Inserts a new scene into the stack
 */
void scenestack_push(scene_t *scn)
{
    logfile_message("scenestack_push(%p)", scn);
    scenestack[scenestack_size++] = scn;
    scn->init();
    logfile_message("scenestack_push() ok");
}



/*
 * scenestack_pop()
 * Deletes the top-most scene of the stack.
 * Please use "return" after calling pop()
 * inside a scene, otherwise the program
 * may crash.
 */
void scenestack_pop()
{
    logfile_message("scenestack_pop()");
    scenestack[scenestack_size-1]->release();
    scenestack[scenestack_size-1] = NULL;
    scenestack_size--;
    logfile_message("scenestack_pop() ok");
}



/*
 * scenestack_top()
 * Returns the top-most scene of the stack.
 */
scene_t *scenestack_top()
{
    return scenestack[scenestack_size-1];
}


/*
 * scenestack_empty()
 * Is the stack empty?
 */
int scenestack_empty()
{
    return (scenestack_size==0);
}







/* private methods */


/* creates a new scene */
scene_t *scene_create(void (*init_func)(), void (*update_func)(), void (*render_func)(), void (*release_func)())
{
    scene_t *s = mallocx(sizeof *s);
    s->init = init_func;
    s->update = update_func;
    s->render = render_func;
    s->release = release_func;
    return s;
}


/* destroys an existing scene */
void scene_destroy(scene_t *scn)
{
    free(scn);
}
