/*
 * video.c - video manager
 * Copyright (C) 2008-2010  Alexandre Martins < alemartf [.a.t.] gmail [.d.o.t.] com >
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <allegro.h>
#include "global.h"
#include "video.h"
#include "util.h"
#include "brick.h"




/* private data */


/* video manager */
#define FILTER_2XSAI            0
#define FILTER_SUPEREAGLE       1
static image_t *video_buffer;
static int video_smooth;
static int video_resolution;
static int video_fullscreen;
static int video_showfps;
static void fast2x_blit(BITMAP *src, BITMAP *dest);
static void filter_blit(BITMAP *src, BITMAP *dest, int filter);
static void window_switch_in();
static void window_switch_out();
static int window_active = TRUE;
static void losangle(image_t *img, int x, int y, int r, int col);

/* Fade-in & fade-out */
#define FADEFX_NONE            0
#define FADEFX_IN              1
#define FADEFX_OUT             2
static int fadefx_type;
static int fadefx_end;
static uint32 fadefx_color;
static double fadefx_elapsed_time;
static double fadefx_total_time;

/* Video Message */
#define VIDEOMSG_TIMEOUT    5000
static uint32 videomsg_endtime;
static char videomsg_data[512];









/* fade effects */

/*
 * fadefx_in()
 * Fade-in effect
 */
void fadefx_in(uint32 color, double seconds)
{
    if(fadefx_type == FADEFX_NONE) {
        fadefx_type = FADEFX_IN;
        fadefx_end = FALSE;
        fadefx_color = color;
        fadefx_elapsed_time = 0;
        fadefx_total_time = seconds;
    }
}


/*
 * fadefx_out()
 * Fade-out effect
 */
void fadefx_out(uint32 color, double seconds)
{
    if(fadefx_type == FADEFX_NONE) {
        fadefx_type = FADEFX_OUT;
        fadefx_end = FALSE;
        fadefx_color = color;
        fadefx_elapsed_time = 0;
        fadefx_total_time = seconds;
    }
}



/*
 * fadefx_over()
 * Asks if the fade effect has ended
 * (only one action when this event loops)
 */
int fadefx_over()
{
    return fadefx_end;
}


/*
 * fadefx_is_fading()
 * Is the fade effect ocurring?
 */
int fadefx_is_fading()
{
    return (fadefx_type != FADEFX_NONE);
}












/* video manager */

/*
 * video_init()
 * Initializes the video manager
 */
void video_init(int resolution, int smooth, int fullscreen, int bpp)
{
    char window_title[128];

    /* video init */
    logfile_message("video_init()");
    set_color_depth(bpp);
    video_buffer = NULL;
    video_changemode(resolution, smooth, fullscreen);

    /* window properties */
    LOCK_FUNCTION(game_quit);
    set_close_button_callback(game_quit);
    sprintf(window_title, GAME_WINDOWTITLE, GAME_TITLE, GAME_VERSION, GAME_SUB_VERSION, GAME_WIP_VERSION);
    set_window_title(window_title);

    /* window callbacks */
    window_active = TRUE;
    if(set_display_switch_mode(SWITCH_BACKGROUND) == 0) {
        if(set_display_switch_callback(SWITCH_IN, window_switch_in) != 0)
            logfile_message("can't set_display_switch_callback(SWTICH_IN, window_switch_in)");

        if(set_display_switch_callback(SWITCH_OUT, window_switch_out) != 0)
            logfile_message("can't set_display_switch_callback(SWTICH_OUT, window_switch_out)");
    }
    else
        logfile_message("can't set_display_switch_mode(SWITCH_BACKGROUND)");

    /* video message */
    videomsg_endtime = 0;
}

/*
 * video_changemode()
 * Sets up the game window
 */
void video_changemode(int resolution, int smooth, int fullscreen)
{
    int width, height;
    int mode;

    logfile_message("video_changemode(%d,%d,%d)", resolution, smooth, fullscreen);

    /* resolution */
    video_resolution = resolution;

    /* fullscreen */
    video_fullscreen = fullscreen;

    /* smooth graphics? */
    video_smooth = smooth;
    if(video_smooth) {
        if(video_get_color_depth() == 8) {
            logfile_message("can't use smooth graphics in the 256 color mode (8 bpp)");
            video_smooth = FALSE;
        }
        else if(video_resolution == VIDEORESOLUTION_1X || video_resolution == VIDEORESOLUTION_EDT) {
            logfile_message("can't use smooth graphics using resolution %d", video_resolution);
            video_smooth = FALSE;
        }
        else
            Init_2xSaI(video_get_color_depth());
    }

    /* creating the backbuffer... */
    if(video_buffer != NULL)
        image_destroy(video_buffer);
    video_buffer = image_create(VIDEO_SCREEN_W, VIDEO_SCREEN_H);
    image_clear(video_get_backbuffer(), image_rgb(0,0,0));

    /* setting up the window... */
# if 0 //ZX
    mode = video_fullscreen ? GFX_AUTODETECT : GFX_AUTODETECT_WINDOWED;
    width = (int)(video_get_window_size().x);
    height = (int)(video_get_window_size().y);
    if(set_gfx_mode(mode, width, height, 0, 0) < 0)
        fatal_error("video_changemode(): couldn't set the graphic mode!\n%s", allegro_error);
# else
    if (video_fullscreen) {
      mode = GFX_AUTODETECT;
      width = 800;
      height = 480;
    } else {
      mode = GFX_AUTODETECT_WINDOWED;
      width = (int)(video_get_window_size().x);
      height = (int)(video_get_window_size().y);
   }
    if(set_gfx_mode(mode, width, height, 0, 0) < 0)
        fatal_error("video_changemode(): couldn't set the graphic mode!\n%s", allegro_error);
# endif
    set_color_conversion(COLORCONV_TOTAL | COLORCONV_KEEP_TRANS);

    logfile_message("video_changemode() ok");
}


/*
 * video_get_resolution()
 * Returns the current resolution value,
 * i.e., VIDEORESOLUTION_*
 */
int video_get_resolution()
{
    return video_resolution;
}


/*
 * video_is_smooth()
 * Smooth graphics?
 */
int video_is_smooth()
{
    return video_smooth;
}


/*
 * video_is_fullscreen()
 * Fullscreen mode?
 */
int video_is_fullscreen()
{
    return video_fullscreen;
}


/*
 * video_get_window_size()
 * Returns the window size, based on
 * the current resolution
 */
v2d_t video_get_window_size()
{
    int width=VIDEO_SCREEN_W, height=VIDEO_SCREEN_H;
    int dw, dh; /* desktop resolution */

    switch(video_resolution) {
        case VIDEORESOLUTION_1X:
            width = VIDEO_SCREEN_W;
            height = VIDEO_SCREEN_H;
            break;

        case VIDEORESOLUTION_2X:
            width = 2*VIDEO_SCREEN_W;
            height = 2*VIDEO_SCREEN_H;
            break;

        case VIDEORESOLUTION_MAX:
            if(get_desktop_resolution(&dw, &dh) == 0) {
                int scale = min((int)(dw/VIDEO_SCREEN_W), (int)(dh/VIDEO_SCREEN_H));
                width = scale*VIDEO_SCREEN_W;
                height = scale*VIDEO_SCREEN_H;
            }
            else {
                width = VIDEO_SCREEN_W;
                height = VIDEO_SCREEN_H;
            }
            break;

        case VIDEORESOLUTION_EDT:
            width = VIDEO_SCREEN_W;
            height = VIDEO_SCREEN_H;
            break;

        default:
            fatal_error("video_get_window_size(): unknown resolution!");
            break;
    }

    return v2d_new(width, height);
}


/*
 * video_get_backbuffer()
 * Returns a pointer to the backbuffer
 */
image_t* video_get_backbuffer()
{
    if(video_buffer == NULL)
        fatal_error("FATAL ERROR: video_get_backbuffer() returned NULL!");

    return video_buffer;
}

/*
 * video_render()
 * Updates the video manager and the screen
 */
void video_render()
{
    /* fade effect */
    fadefx_end = FALSE;
    if(fadefx_type != FADEFX_NONE) {
        fadefx_elapsed_time += game_delta();
        if(fadefx_elapsed_time < fadefx_total_time) {
            if(video_get_color_depth() > 8) {
                /* true-color fade effect */
                int n;

                n = (int)( (double)255 * (fadefx_elapsed_time*1.25 / fadefx_total_time) );
                n = clip(n, 0, 255);
                n = (fadefx_type == FADEFX_IN) ? 255-n : n;

                drawing_mode(DRAW_MODE_TRANS, NULL, 0, 0);
                set_trans_blender(0, 0, 0, n);
                rectfill(video_get_backbuffer()->data, 0, 0, VIDEO_SCREEN_W, VIDEO_SCREEN_H, fadefx_color);
                solid_mode();
            }
            else {
                /* 256-color fade effect */
                int i, j, x, y, r;
                double prob;

                prob = fadefx_elapsed_time / fadefx_total_time;
                prob = (fadefx_type == FADEFX_IN) ? 1.0 - prob : prob;

                for(i=0; i<=20; i++) {
                    for(j=0; j<=10; j++) {
                        r = (int)( ((1.0-(double)i/20)/8 + 7*prob/8)*50.0 );
                        x = (int)((double)i/20 * video_get_backbuffer()->w);
                        y = (int)((double)j/10 * video_get_backbuffer()->h);
                        losangle(video_get_backbuffer(), x, y, r, fadefx_color);
                    }
                }
            }
        }
        else {
            if(fadefx_type == FADEFX_OUT)
                rectfill(video_get_backbuffer()->data, 0, 0, VIDEO_SCREEN_W, VIDEO_SCREEN_H, fadefx_color);
            fadefx_type = FADEFX_NONE;
            fadefx_total_time = fadefx_elapsed_time = 0;
            fadefx_color = 0;
            fadefx_end = TRUE;
        }
    }



    /* video message */
    if(game_timer() < videomsg_endtime)
        textout_ex(video_get_backbuffer()->data, font, videomsg_data, 0, VIDEO_SCREEN_H-text_height(font), makecol(255,255,255), makecol(0,0,0));


    /* fps counter */
    if(video_is_fps_visible())
        textprintf_right_ex(video_get_backbuffer()->data, font, VIDEO_SCREEN_W, 0, makecol(255,255,255), makecol(0,0,0),"FPS: %d", game_fps());


    /* render */
    switch(video_get_resolution()) {
        /* tiny window */
        case VIDEORESOLUTION_1X:
        {
            blit(video_get_backbuffer()->data, screen, 0, 0, 0, 0, VIDEO_SCREEN_W, VIDEO_SCREEN_H);
            break;
        }

        /* double size */
        case VIDEORESOLUTION_2X:
        {
            BITMAP *tmp = create_bitmap((int)(video_get_window_size().x), (int)(video_get_window_size().y));

            if(video_is_smooth())
                filter_blit(video_get_backbuffer()->data, tmp, FILTER_2XSAI);
            else
                fast2x_blit(video_get_backbuffer()->data, tmp);

            blit(tmp, screen, 0, 0, 0, 0, VIDEO_SCREEN_W*2, VIDEO_SCREEN_H*2);
            destroy_bitmap(tmp);
            break;
        }

        /* maximum size */
        case VIDEORESOLUTION_MAX:
        {
# ifdef PND_MODE
            BITMAP *tmp = create_bitmap(800, 480 );
# else
            BITMAP *tmp = create_bitmap((int)(video_get_window_size().x), (int)(video_get_window_size().y));
# endif

            if(video_is_smooth() && tmp->w >= 2*VIDEO_SCREEN_W && tmp->h >= 2*VIDEO_SCREEN_H) {
                BITMAP *half = create_bitmap(tmp->w/2, tmp->h/2);
                stretch_blit(video_get_backbuffer()->data, half, 0, 0, VIDEO_SCREEN_W, VIDEO_SCREEN_H, 0, 0, half->w, half->h);
                filter_blit(half, tmp, FILTER_2XSAI);
                destroy_bitmap(half);
            }
            else {
                stretch_blit(video_get_backbuffer()->data, tmp, 0, 0, VIDEO_SCREEN_W, VIDEO_SCREEN_H, 0, 0, tmp->w, tmp->h);
            }
            blit(tmp, screen, 0, 0, 0, 0, tmp->w, tmp->h);
            destroy_bitmap(tmp);
            break;
        }

        /* editor */
        case VIDEORESOLUTION_EDT:
        {
            blit(video_get_backbuffer()->data, screen, 0, 0, 0, 0, VIDEO_SCREEN_W, VIDEO_SCREEN_H);
            break;
        }
    }
}


/*
 * video_release()
 * Releases the video manager
 */
void video_release()
{
    logfile_message("video_release()");
    image_destroy(video_buffer);
}


/*
 * video_showmessage()
 * Shows a text message to the user
 */
void video_showmessage(char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    vsprintf(videomsg_data, fmt, args);
    va_end(args);

    videomsg_endtime = game_timer() + VIDEOMSG_TIMEOUT;
}


/*
 * video_get_color_depth()
 * Returns the current color depth
 */
int video_get_color_depth()
{
    return get_color_depth();
}


/*
 * video_is_window_active()
 * Returns TRUE if the game window is active,
 * or FALSE otherwise
 */
int video_is_window_active()
{
    return window_active;
}



/*
 * video_get_maskcolor()
 * Returns the mask color
 */
uint32 video_get_maskcolor()
{
    switch(video_get_color_depth()) {
        case 8:  return MASK_COLOR_8;
        case 16: return MASK_COLOR_16;
        case 24: return MASK_COLOR_24;
        case 32: return MASK_COLOR_32;
    }

    return MASK_COLOR_16;
}


/*
 * video_show_fps()
 * Shows/hides the FPS counter
 */
void video_show_fps(int show)
{
    video_showfps = show;
}


/*
 * video_is_fps_visible()
 * Is the FPS counter visible?
 */
int video_is_fps_visible()
{
    return video_showfps;
}









/* images */

/*
 * image_load()
 * Loads a image from a file.
 * Supported types: PNG, BMP, PCX, TGA
 */
image_t *image_load(char *path)
{
    image_t *img;
    char abs_path[1024];

    resource_filepath(abs_path, path, sizeof(abs_path), RESFP_READ);
    logfile_message("image_load(%s)", abs_path);

    if(NULL == (img=(image_t*)hash_search(path, DATATYPE_IMAGE))) {
        img = mallocx(sizeof *img);    
        img->data = load_bitmap(abs_path, NULL);
        if(img->data) {
            img->w = img->data->w;
            img->h = img->data->h;
            hash_insert(path, (void*)img, DATATYPE_IMAGE); /* do NOT use abs_path in the hash table! */
            hash_ref(path, DATATYPE_IMAGE);
            logfile_message("image_load() ok");
        }
        else {
            logfile_message("image_load() error! %s", allegro_error);
            free(img);
            img = NULL;
        }
    }
    else
        hash_ref(path, DATATYPE_IMAGE);

    return img;
}



/*
 * image_unref()
 * Used for reference counting. Normally you
 * don't need to bother with this, unless
 * you care about reducing memory usage.
 * Note that 'image_ref()' must not exist.
 * Returns the no. of references to the image
 */
int image_unref(char *path)
{
    return hash_unref(path, DATATYPE_IMAGE);
}



/*
 * image_save()
 * Saves a image to a file
 */
void image_save(image_t *img, char *path)
{
    char abs_path[1024];

    resource_filepath(abs_path, path, sizeof(abs_path), RESFP_WRITE);
    logfile_message("image_save(%p,%s)", img, abs_path);

    if(video_get_color_depth() > 8)
        save_bitmap(abs_path, img->data, NULL);
    else {
        PALETTE pal;
        get_palette(pal);
        save_bitmap(abs_path, img->data, pal);
    }
}



/*
 * image_create()
 * Creates a new image of a given size
 */
image_t *image_create(int width, int height)
{
    image_t *img = mallocx(sizeof *img);

    img->data = create_bitmap(width, height);
    img->w = width;
    img->h = height;
    image_clear(img, image_rgb(0,0,0));

    return img;
}


/*
 * image_destroy()
 * Destroys an image
 */
void image_destroy(image_t *img)
{
    destroy_bitmap(img->data);
    free(img);
}


/*
 * image_getpixel()
 * Returns the pixel at the given position on the image
 */
uint32 image_getpixel(image_t *img, int x, int y)
{
    return getpixel(img->data, x, y);
}


/*
 * image_putpixel()
 * Plots a pixel into the given image
 */
void image_putpixel(image_t *img, int x, int y, uint32 color)
{
    putpixel(img->data, x, y, color);
}

/*
 * image_rgb()
 * Generates an uint32 color
 */
uint32 image_rgb(uint8 r, uint8 g, uint8 b)
{
    return makecol(r,g,b);
}


/*
 * image_color2rgb()
 * Converts an uint32 to a (r,g,b) triple
 */
void image_color2rgb(uint32 color, uint8 *r, uint8 *g, uint8 *b)
{
    if(r) *r = getr(color);
    if(g) *g = getg(color);
    if(b) *b = getb(color);
}


/*
 * image_clear()
 * Clears an given image with some color
 */
void image_clear(image_t *img, uint32 color)
{
    clear_to_color(img->data, color);
}


/*
 * image_blit()
 * Blits a surface onto another
 */
void image_blit(image_t *src, image_t *dest, int source_x, int source_y, int dest_x, int dest_y, int width, int height)
{
    blit(src->data, dest->data, source_x, source_y, dest_x, dest_y, width, height);
}


/*
 * image_draw()
 * Draws an image onto the destination surface
 * at the specified position
 *
 * flags: refer to the IF_* defines (Image Flags)
 */
void image_draw(image_t *src, image_t *dest, int x, int y, uint32 flags)
{
    if((flags & IF_HFLIP) && !(flags & IF_VFLIP))
        draw_sprite_h_flip(dest->data, src->data, x, y);
    else if(!(flags & IF_HFLIP) && (flags & IF_VFLIP))
        draw_sprite_v_flip(dest->data, src->data, x, y);
    else if((flags & IF_HFLIP) && (flags & IF_VFLIP))
        draw_sprite_vh_flip(dest->data, src->data, x, y);
    else
        draw_sprite(dest->data, src->data, x, y);
}



/*
 * image_draw_scaled()
 * Draws a scaled image onto the destination surface
 * at the specified position
 *
 * scale: (1.0, 1.0) is the original size
 *        (2.0, 2.0) stands for a double-sized image
 *        (0.5, 0.5) stands for a smaller image
 */
void image_draw_scaled(image_t *src, image_t *dest, int x, int y, v2d_t scale, uint32 flags)
{
    image_t *tmp;
    int w = (int)(scale.x * src->w);
    int h = (int)(scale.y * src->h);

    tmp = image_create(w, h);
    stretch_blit(src->data, tmp->data, 0, 0, src->w, src->h, 0, 0, w, h);
    image_draw(tmp, dest, x, y, flags);
    image_destroy(tmp);
}


/*
 * image_draw_rotated()
 * Draws a rotated image onto the destination bitmap at the specified position 
 *
 * ang: angle given in radians
 * cx, cy: pivot positions
 */
void image_draw_rotated(image_t *src, image_t *dest, int x, int y, int cx, int cy, double ang, uint32 flags)
{
    double conv = (-ang * (180.0/PI)) * (64.0/90.0);

    if((flags & IF_HFLIP) && !(flags & IF_VFLIP))
        pivot_sprite_v_flip(dest->data, src->data, x, y, cx, 0, ftofix((float)(conv+128)));
    else if(!(flags & IF_HFLIP) && (flags & IF_VFLIP))
        pivot_sprite_v_flip(dest->data, src->data, x, y, 0, cy, ftofix((float)conv));
    else if((flags & IF_HFLIP) && (flags & IF_VFLIP))
        pivot_sprite(dest->data, src->data, x, y, 0, 0, ftofix((float)(conv+128)));
    else
        pivot_sprite(dest->data, src->data, x, y, cx, cy, ftofix((float)conv));
}

 
/*
 * image_draw_trans()
 * Draws a translucent image
 *
 * color: lit truecolor pixels
 * alpha: 0.0 (invisible) <= alpha <= 1.0 (opaque)
 */
void image_draw_trans(image_t *src, image_t *dest, int x, int y, uint32 color, double alpha, uint32 flags)
{
    image_t *tmp;
    uint8 r, g, b;

    alpha = clip(alpha, 0.0, 1.0);
    image_color2rgb(color, &r, &g, &b);
    set_trans_blender(r, g, b, (int)(255 * alpha));

    tmp = image_create(src->w, src->h);
    image_clear(tmp, video_get_maskcolor());
    image_draw(src, tmp, 0, 0, flags);
    draw_trans_sprite(dest->data, tmp->data, x, y);
    image_destroy(tmp);
}
 





/* private stuff */

/* filter_blit() applies the graphic filter
 * filter, fixing the possible defects of the
 * resulting image.
 *
 * if (filter == 2xsai) or (filter == superagle):
 * -- we assume that:
 * ---- dest->w = 2 * src->w
 * ---- dest->h = 2 * src->h
 */
void filter_blit(BITMAP *src, BITMAP *dest, int filter)
{
    int i, j, k=2;

    switch(filter) {
        case FILTER_2XSAI:
            Super2xSaI(src, dest, 0, 0, 0, 0, src->w, src->h);
            for(i=0; i<dest->h; i++) { /* image fix */
                for(j=0; j<k; j++)
                    putpixel(dest, j, i, getpixel(dest, k, i));
            }
            break;

        case FILTER_SUPEREAGLE:
            SuperEagle(src, dest, 0, 0, 0, 0, src->w, src->h);
            for(i=0; i<dest->h; i++) { /* image fix */
                for(j=0; j<k; j++)
                    putpixel(dest, dest->w-1-j, i, getpixel(dest, dest->w-1-k, i));
            }
            break;
    }
}

/* fast2x_blit resizes the src image by a
 * factor of 2. It assumes that:
 *
 * src is a memory bitmap
 * dest is a previously created memory bitmap
 * dest->w == 2 * src->w
 * dest->h == 2 * src->h */
void fast2x_blit(BITMAP *src, BITMAP *dest)
{
    int i, j;

    switch(video_get_color_depth())
    {
        case 8:
            for(j=0; j<dest->h; j++) {
                for(i=0; i<dest->w; i++)
                    ((uint8*)dest->line[j])[i] = ((uint8*)src->line[j/2])[i/2];
            }
            break;

        case 16:
            for(j=0; j<dest->h; j++) {
                for(i=0; i<dest->w; i++)
                    ((uint16*)dest->line[j])[i] = ((uint16*)src->line[j/2])[i/2];
            }
            break;

        case 24:
            /* TODO */
            stretch_blit(src, dest, 0, 0, src->w, src->h, 0, 0, dest->w, dest->h);
            break;

        case 32:
            for(j=0; j<dest->h; j++) {
                for(i=0; i<dest->w; i++)
                    ((uint32*)dest->line[j])[i] = ((uint32*)src->line[j/2])[i/2];
            }
            break;

        default:
            break;
    }
}

/* this window is active */
void window_switch_in()
{
    window_active = TRUE;
}


/* this window is not active */
void window_switch_out()
{
    window_active = FALSE;
}

/* losangle */
void losangle(image_t *img, int x, int y, int r, int col)
{
    int points[] = {
        x,   y-r,
        x-r, y,
        x,   y+r,
        x+r, y
    };

    polygon(img->data, 4, points, col);
}
